#!/usr/bin/env python3
# vim: set fileencoding=utf-8

import iio

from osgeo import gdal
import glob
import os
import argparse
import math
import sys, getopt
from scipy.sparse import eye, kron, diags, vstack
from scipy.sparse.linalg import spsolve 
from scipy import ndimage
from numpy.random import randn, rand
import numpy as np
import numpy.matlib
from scikits.umfpack import spsolve, splu

def get_az_el(i):
    
    ds = gdal.Open(i)
    meta = ds.GetMetadata()
    keys = list(k.lower() for k in meta)
    combined = '\t'.join(keys)
    
    if (('SUN_AZ').lower() not in combined) or (('SUN_EL').lower() not in combined):
        sys.exit("Sun azimuth and elevation not found in metadata. "
             "Provide sun informations and try again.")
        
    
    az = [value for key, value in meta.items() if 'SUN_AZ' in key.upper()]
    el = [value for key, value in meta.items() if 'SUN_EL' in key.upper()]
    return np.array([az, el])


def sun_direction_image(imtif, h, w, az_el):
    az = az_el[0]
    el = az_el[1]
    print(az, el)
    azrad = float(az) * math.pi / 180
    elrad = float(el) * math.pi / 180
    nx = math.cos(elrad) * math.sin (azrad)
    ny = - math.cos(elrad) * math.cos (azrad)
    
    xx, yy = np.meshgrid(np.linspace(0, w-1, w), np.linspace(0, h-1, h))
    sun =  - nx * xx - ny * yy
    return sun



def sfs(s, d, m, o):
    if (len(s.shape) == 3) :
        s = np.sum(s, axis=2)
    #h, w, _ = s.shape
    #s = s[:,:,0]+s[:,:,1]+s[:,:,2]
    h, w = s.shape
    
    m = (m[:,:,0] > 0).astype(float)
    
    s2 = np.copy(s)
    s2[m == 0] = np.nan
    s3 = (s-np.nanmean(s2[:]))/np.nanstd(s2[:])
    s4 = s3 - 0
    s4[m == 0] = np.nan
    
    mask = ndimage.binary_dilation((s3*m)>=(np.nanmean(s4[:]) + 4), np.ones((6,6)))*m
    s5 = s - 0
    s5[mask > 0] = np.nan
    s5[m == 0] = np.nan
    s = (s-np.nanmean(s5.flatten()))/np.nanstd(s5.flatten());
    
    B1 = kron(eye(h), eye(w,w,1) - eye(w))
    B2 = kron(eye(h,h,1) - eye(h), eye(w))
    B = vstack([B1, B2])
    
    t = ((B @ mask.flatten()) != 0)
    ind = np.array((np.logical_not(t)).nonzero())
    
    Bcsr = B.tocsr()
    Bcsr = Bcsr[ind.flatten(), :]
    B = Bcsr.tocoo()
    C = abs(B)/2;
    L = -B.T @ B
    
    A1 = C.T @ diags(B @ d.flatten()) @ B ;
    
    I = eye(w * h)
    M = diags(0 + (m.flatten() > 0)*(1-mask.flatten()))
    e = 0.7
    
    A = A1.T @ M @ A1 - e*L + I - M;
    b = A1.T @ M @ s.flatten()
    
    u = spsolve(A, b)
    iio.write(o, u.reshape(h,w))
    quantize(u.reshape(h,w), "output.png")

def main(i, m, o, tf, c=False, a=0.0, e=0.0):
    im = iio.read(i)
    mask = iio.read(m)
    h,w,_ = im.shape
    t = open(tf,"w")
    if c:
        az_el = np.array([a, e])
        t.write('User given light source direction \n az: ' + str(a) + ' el: ' + str(e))
    else:
        az_el = get_az_el(i)
        t.write('Light source direction found in metadata \n az: ' + str(az_el[0]) + ' el: ' + str(az_el[1]))

    t.close()
    d = sun_direction_image(i, h, w, az_el)
    sfs(im, d, mask, o)
    quantize(im, "input.png")


def quantize(i, f):
    # requantification in [0, 255]

#    mini = np.min(im)
#    maxi = np.max(im)
    mini = np.percentile(i.flatten(), 2)
    maxi = np.percentile(i.flatten(), 98)

    im = 255.0 * (i - mini) / (maxi - mini)
    im[im < 0] = 0
    im[im > 255] = 255

    if (len(im.shape) < 3):
        iio.write(f, im)
    else:
        if (im.shape[2] >= 3):
            iio.write(f, im[:,:,0:3])
        else:
            iio.write(f,im)


def cli():
    """
    Command line interface
    """
    # Program description
    parser = argparse.ArgumentParser(description='Stabilize contrast')
    # Mandatory parameters
    parser.add_argument('-i', '--im_file', required=True, type=str,
                        help='colour stabilised images directory')
    parser.add_argument('-o', '--out_file', required=True, type=str,
                        help='output directory (created if nonexistent)')
    parser.add_argument('-m', '--mname', required=True, type=str,
                        help='mask path (1 in the ROI, 0 outside)')
    parser.add_argument('-t', '--texte', required=True, type=str,
                        help='mask path (1 in the ROI, 0 outside)')
    parser.add_argument('-c', '--check', action='store_true', default=False,
                        help='mask path (1 in the ROI, 0 outside)')
    parser.add_argument('-a', '--azimuth', type=float, default=0.0,
                        help='azimuth')
    parser.add_argument('-e', '--elevation', type=float, default=0.0,
                        help='elevation')

    args = parser.parse_args()

    main(args.im_file, args.mname, args.out_file, args.texte, args.check, args.azimuth, args.elevation)

if __name__ == '__main__':
    cli()

