#!/usr/bin/env python3
# vim: set fileencoding=utf-8

import iio

from osgeo import gdal

import time
import glob
import os

import math

import sys, getopt

from scipy.sparse import eye, kron, diags, vstack
from scipy.sparse.linalg import spsolve
from scipy import ndimage

from numpy.random import randn, rand
import numpy as np
import numpy.matlib

from scikits.umfpack import spsolve, splu


def get_az_el(i, az_el):
    az = az_el[0]
    el = az_el[1]
    print(type(az))
    if (az < 360) and (el < 360) and (az >= 0) and (el >= 0):
        print("Use azimuth and elevation provided by the user.")
        return np.array([az, el])
    if (az == 360) or (el == 360):
        print("Sun azimuth and elevation not provided, "
             "searching in image metadata.")
    elif (az > 360) or (el > 360) or (az < 0) or (el < 0):
        print("Incorrect sun azimuth and elevation (expecting values between 0 and 359), "
             "searching in image metadata.")
    print("1", az, el)
    
    ds = gdal.Open(i)
    meta = ds.GetMetadata()
    keys = list(k.lower() for k in meta)
    combined = '\t'.join(keys)
    
    if (('SUN_AZ').lower() not in combined) or (('SUN_EL').lower() not in combined):
        sys.exit("Sun azimuth and elevation not found in metadata. "
             "Provide sun informations and try again.")
        
    
    az = [value for key, value in meta.items() if 'SUN_AZ' in key.upper()]
    el = [value for key, value in meta.items() if 'SUN_EL' in key.upper()]
    print("2", az, el)
    return np.array([az, el])


def sun_direction_image(imtif, h, w, az_el):
    az = az_el[0]
    el = az_el[1]
    print(az, el)
    azrad = float(az) * math.pi / 180
    elrad = float(el) * math.pi / 180
    nx = math.cos(elrad) * math.sin (azrad)
    ny = - math.cos(elrad) * math.cos (azrad)
    
    xx, yy = np.meshgrid(np.linspace(0, w-1, w), np.linspace(0, h-1, h))
    sun =  - nx * xx - ny * yy
    return sun



def sfs(s, d, m, o):
    h, w, _ = s.shape
    s = s[:,:,0]+s[:,:,1]+s[:,:,2]
    
    m = (m[:,:,0] > 0).astype(float)
    
    s2 = np.copy(s)
    s2[m == 0] = np.nan
    s3 = (s-np.nanmean(s2[:]))/np.nanstd(s2[:])
    s4 = s3 - 0
    s4[m == 0] = np.nan
    
    mask = ndimage.binary_dilation((s3*m)>=(np.nanmean(s4[:]) + 4), np.ones((6,6)))*m
    s5 = s - 0
    s5[mask > 0] = np.nan
    s5[m == 0] = np.nan
    s = (s-np.nanmean(s5.flatten()))/np.nanstd(s5.flatten());
    
    B1 = kron(eye(h), eye(w,w,1) - eye(w))
    B2 = kron(eye(h,h,1) - eye(h), eye(w))
    B = vstack([B1, B2])
    
    t = ((B @ mask.flatten()) != 0)
    ind = np.array((np.logical_not(t)).nonzero())
    
    Bcsr = B.tocsr()
    Bcsr = Bcsr[ind.flatten(), :]
    B = Bcsr.tocoo()
    C = abs(B)/2;
    L = -B.T @ B
    
    A1 = C.T @ diags(B @ d.flatten()) @ B ;
    
    I = eye(w * h)
    M = diags(0 + (m.flatten() > 0)*(1-mask.flatten()))
    e = 0.7
    
    A = A1.T @ M @ A1 - e*L + I - M;
    b = A1.T @ M @ s.flatten()
    
    factor = cholesky(A)
    u = factor(b)
    iio.write(o, u.reshape(h,w))

def main(i, o, m, azel=np.array([360.0, 360.0])):
    im = iio.read(i)
    mask = iio.read(m)
    h,w,_ = im.shape
    az_el = get_az_el(i, azel)
    d = sun_direction_image(i, h, w, az_el)
    sfs(im, d, mask, o)



def cli():
    """ 
    Command line interface
    """
    # Program description
    parser = argparse.ArgumentParser(description='Stabilize contrast')
    # Mandatory parameters
    parser.add_argument('-i', '--im', required=True, type=str,
                        help='input image')
    parser.add_argument('-o', '--out', required=True, type=str,
                        help='output image')
    parser.add_argument('-m', '--mask', required=True, type=str,
                        help='mask (>0 in the ROI)')
    # Optional parameters
    parser.add_argument('--verbose', action='store_true', default=False,
                        help='print non-critical information.')
    
    # Note: argparse infers var name from arg name and replaces any '-' by '_'
    args = parser.parse_args()
    
    main(args.im, args.out, args.mask)
    
if __name__ == '__main__':
    cli()
    
    
    
    
    
    
    
